"""
This module defines interfaces for requests and responses used in Connexion for authentication,
validation, serialization, etc.
"""
import typing as t
from collections import defaultdict

from python_multipart.multipart import parse_options_header
from starlette.datastructures import UploadFile
from starlette.requests import Request as StarletteRequest
from werkzeug import Request as WerkzeugRequest

from connexion.http_facts import FORM_CONTENT_TYPES
from connexion.utils import is_json_mimetype


class _RequestInterface:
    @property
    def context(self) -> t.Dict[str, t.Any]:
        """The connexion context of the current request cycle."""
        raise NotImplementedError

    @property
    def content_type(self) -> str:
        """The content type included in the request headers."""
        raise NotImplementedError

    @property
    def mimetype(self) -> str:
        """The content type included in the request headers stripped from any optional character
        set encoding"""
        raise NotImplementedError

    @property
    def path_params(self) -> t.Dict[str, t.Any]:
        """Path parameters exposed as a dictionary"""
        raise NotImplementedError

    @property
    def query_params(self) -> t.Dict[str, t.Any]:
        """Query parameters exposed as a dictionary"""
        raise NotImplementedError

    def form(self) -> t.Union[t.Dict[str, t.Any], t.Awaitable[t.Dict[str, t.Any]]]:
        """Form data, including files."""
        raise NotImplementedError

    def files(self) -> t.Dict[str, t.Any]:
        """Files included in the request."""
        raise NotImplementedError

    def json(self) -> dict:
        """Json data included in the request."""
        raise NotImplementedError

    def get_body(self) -> t.Any:
        """Get body based on the content type. This returns json data for json content types,
        form data for form content types, and bytes for all others. If the bytes data is empty,
        :code:`None` is returned instead."""
        raise NotImplementedError


class WSGIRequest(_RequestInterface):
    def __init__(
        self, werkzeug_request: WerkzeugRequest, uri_parser=None, view_args=None
    ):
        self._werkzeug_request = werkzeug_request
        self.uri_parser = uri_parser
        self.view_args = view_args

        self._context = None
        self._path_params = None
        self._query_params = None
        self._form = None
        self._body = None

    @property
    def context(self):
        if self._context is None:
            scope = self.environ["asgi.scope"]
            extensions = scope.setdefault("extensions", {})
            self._context = extensions.setdefault("connexion_context", {})
        return self._context

    @property
    def content_type(self) -> str:
        return self._werkzeug_request.content_type or "application/octet-stream"

    @property
    def mimetype(self) -> str:
        return self._werkzeug_request.mimetype

    @property
    def path_params(self):
        if self._path_params is None:
            self._path_params = self.uri_parser.resolve_path(self.view_args)
        return self._path_params

    @property
    def query_params(self):
        if self._query_params is None:
            query_params = {k: self.args.getlist(k) for k in self.args}
            self._query_params = self.uri_parser.resolve_query(query_params)
        return self._query_params

    def form(self):
        if self._form is None:
            form = self._werkzeug_request.form.to_dict(flat=False)
            self._form = self.uri_parser.resolve_form(form)
        return self._form

    def files(self):
        return self._werkzeug_request.files.to_dict(flat=False)

    def json(self):
        return self.get_json(silent=True)

    def get_body(self):
        if self._body is None:
            if is_json_mimetype(self.content_type):
                self._body = self.get_json(silent=True)
            elif self.mimetype in FORM_CONTENT_TYPES:
                self._body = self.form()
            else:
                # Return explicit None instead of empty bytestring so it is handled as null downstream
                self._body = self.get_data() or None
        return self._body

    def __getattr__(self, item):
        return getattr(self._werkzeug_request, item)


class ConnexionRequest(_RequestInterface):
    """
    Implementation of the Connexion :code:`_RequestInterface` representing an ASGI request.

    .. attribute:: _starlette_request
        :noindex:

        This class wraps a Starlette `Request <https://www.starlette.io/requests/#request>`_,
        and provides access to its attributes by proxy.

    """

    def __init__(self, *args, uri_parser=None, **kwargs):
        # Might be set in `from_starlette_request` class method
        if not hasattr(self, "_starlette_request"):
            self._starlette_request = StarletteRequest(*args, **kwargs)
        self.uri_parser = uri_parser

        self._context = None
        self._mimetype = None
        self._path_params = None
        self._query_params = None
        self._form = None
        self._files = None

    @classmethod
    def from_starlette_request(
        cls, request: StarletteRequest, uri_parser=None
    ) -> "ConnexionRequest":
        # Instantiate the class, and set the `_starlette_request` property before initializing.
        self = cls.__new__(cls)
        self._starlette_request = request
        self.__init__(uri_parser=uri_parser)  # type: ignore
        return self

    @property
    def context(self):
        if self._context is None:
            extensions = self.scope.setdefault("extensions", {})
            self._context = extensions.setdefault("connexion_context", {})
        return self._context

    @property
    def content_type(self):
        return self.headers.get("content-type", "application/octet-stream")

    @property
    def mimetype(self):
        if not self._mimetype:
            mimetype, _ = parse_options_header(self.content_type)
            self._mimetype = mimetype.decode()
        return self._mimetype

    @property
    def path_params(self) -> t.Dict[str, t.Any]:
        if self._path_params is None:
            self._path_params = self.uri_parser.resolve_path(
                self._starlette_request.path_params
            )
        return self._path_params

    @property
    def query_params(self):
        if self._query_params is None:
            args = self._starlette_request.query_params
            query_params = {k: args.getlist(k) for k in args}
            self._query_params = self.uri_parser.resolve_query(query_params)
        return self._query_params

    async def form(self):
        if self._form is None:
            await self._split_form_files()
        return self._form

    async def files(self):
        if self._files is None:
            await self._split_form_files()
        return self._files

    async def _split_form_files(self):
        form_data = await self._starlette_request.form()

        files = defaultdict(list)
        form = defaultdict(list)
        for k, v in form_data.multi_items():
            if isinstance(v, UploadFile):
                files[k].append(v)
            else:
                form[k].append(v)

        self._files = files
        self._form = self.uri_parser.resolve_form(form)

    async def json(self):
        try:
            return await self._starlette_request.json()
        except ValueError:
            return None

    async def get_body(self):
        if is_json_mimetype(self.content_type):
            return await self.json()
        elif self.mimetype in FORM_CONTENT_TYPES:
            return await self.form()
        else:
            # Return explicit None instead of empty bytestring so it is handled as null downstream
            return await self.body() or None

    def __getattr__(self, item):
        if self.__getattribute__("_starlette_request"):
            return getattr(self._starlette_request, item)


class ConnexionResponse:
    """Connexion interface for a response."""

    def __init__(
        self,
        status_code=200,
        mimetype=None,
        content_type=None,
        body=None,
        headers=None,
        is_streamed=False,
    ):
        self.status_code = status_code
        self.mimetype = mimetype
        self.content_type = content_type
        self.body = body
        self.headers = headers or {}
        if content_type:
            self.headers.update({"Content-Type": content_type})
        self.is_streamed = is_streamed



import logging
import typing as t

from jsonschema import Draft4Validator, ValidationError
from starlette.datastructures import Headers, UploadFile
from starlette.formparsers import FormParser, MultiPartParser
from starlette.types import Scope

from connexion.exceptions import BadRequestProblem, ExtraParameterProblem
from connexion.json_schema import Draft4RequestValidator, format_error_with_path
from connexion.uri_parsing import AbstractURIParser
from connexion.validators import AbstractRequestBodyValidator

logger = logging.getLogger("connexion.validators.form_data")


class FormDataValidator(AbstractRequestBodyValidator):
    """Request body validator for form content types."""

    def __init__(
        self,
        *,
        schema: dict,
        required=False,
        nullable=False,
        encoding: str,
        strict_validation: bool,
        uri_parser: t.Optional[AbstractURIParser] = None,
    ) -> None:
        super().__init__(
            schema=schema,
            required=required,
            nullable=nullable,
            encoding=encoding,
            strict_validation=strict_validation,
        )
        self._uri_parser = uri_parser

    @property
    def _validator(self):
        return Draft4RequestValidator(
            self._schema, format_checker=Draft4Validator.FORMAT_CHECKER
        )

    @property
    def _form_parser_cls(self):
        return FormParser

    async def _parse(self, stream: t.AsyncGenerator[bytes, None], scope: Scope) -> dict:
        headers = Headers(scope=scope)
        form_parser = self._form_parser_cls(headers, stream)
        data = await form_parser.parse()

        if self._uri_parser is not None:
            # Don't parse file_data
            form_data = {}
            file_data: t.Dict[str, t.Union[str, t.List[str]]] = {}
            for key in data.keys():
                # Extract files
                param_schema = self._schema.get("properties", {}).get(key, {})
                value = data.getlist(key)

                def is_file(schema):
                    return schema.get("type") == "string" and schema.get("format") in [
                        "binary",
                        "base64",
                    ]

                # Single file upload
                if is_file(param_schema):
                    # Unpack if single file received
                    if len(value) == 1:
                        file_data[key] = ""
                    # If multiple files received, replace with array so validation will fail
                    else:
                        file_data[key] = [""] * len(value)
                # Multiple file upload, replace files with array of strings
                elif is_file(param_schema.get("items", {})):
                    file_data[key] = [""] * len(value)
                # UploadFile received for non-file upload. Replace and let validation handle.
                elif isinstance(value[0], UploadFile):
                    file_data[key] = [""] * len(value)
                # No files, add multi-value to form data and let uri parser handle multi-value
                else:
                    form_data[key] = value

            # Resolve form data, not file data
            data = self._uri_parser.resolve_form(form_data)
            # Add the files again
            data.update(file_data)
        else:
            data = {k: data.getlist(k) for k in data}

        return data

    def _validate(self, body: t.Any) -> t.Optional[dict]:  # type: ignore[return]
        if not isinstance(body, dict):
            raise BadRequestProblem("Parsed body must be a mapping")
        if self._strict_validation:
            self._validate_params_strictly(body)
        try:
            self._validator.validate(body)
        except ValidationError as exception:
            error_path_msg = format_error_with_path(exception=exception)
            logger.error(
                f"Validation error: {exception.message}{error_path_msg}",
                extra={"validator": "body"},
            )
            raise BadRequestProblem(detail=f"{exception.message}{error_path_msg}")

    def _validate_params_strictly(self, data: dict) -> None:
        form_params = data.keys()
        spec_params = self._schema.get("properties", {}).keys()
        errors = set(form_params).difference(set(spec_params))
        if errors:
            raise ExtraParameterProblem(param_type="formData", extra_params=errors)


class MultiPartFormDataValidator(FormDataValidator):
    @property
    def _form_parser_cls(self):
        return MultiPartParser



"""
This module defines functionality specific to the Flask framework.
"""
import functools
import random
import re
import string
import typing as t

import flask
import werkzeug.routing

from connexion import jsonifier
from connexion.frameworks.abstract import Framework
from connexion.lifecycle import WSGIRequest
from connexion.uri_parsing import AbstractURIParser


class Flask(Framework):
    @staticmethod
    def is_framework_response(response: t.Any) -> bool:
        return isinstance(response, flask.Response) or isinstance(
            response, werkzeug.wrappers.Response
        )

    @classmethod
    def connexion_to_framework_response(cls, response):
        return cls.build_response(
            content_type=response.content_type,
            headers=response.headers,
            status_code=response.status_code,
            data=response.body,
        )

    @classmethod
    def build_response(
        cls,
        data: t.Any,
        *,
        content_type: t.Optional[str] = None,
        headers: t.Optional[dict] = None,
        status_code: int = None
    ):
        if cls.is_framework_response(data):
            return flask.current_app.make_response((data, status_code, headers))

        kwargs = {
            "mimetype": content_type,
            "headers": headers,
            "response": data,
            "status": status_code,
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return flask.current_app.response_class(**kwargs)

    @staticmethod
    def get_request(*, uri_parser: AbstractURIParser, **kwargs) -> WSGIRequest:  # type: ignore
        return WSGIRequest(
            flask.request, uri_parser=uri_parser, view_args=flask.request.view_args
        )


PATH_PARAMETER = re.compile(r"\{([^}]*)\}")

# map Swagger type to flask path converter
# see http://flask.pocoo.org/docs/0.10/api/#url-route-registrations
PATH_PARAMETER_CONVERTERS = {"integer": "int", "number": "float", "path": "path"}


def flaskify_endpoint(identifier, randomize=None):
    """
    Converts the provided identifier in a valid flask endpoint name

    :type identifier: str
    :param randomize: If specified, add this many random characters (upper case
        and digits) to the endpoint name, separated by a pipe character.
    :type randomize: int | None
    :rtype: str

    """
    result = identifier.replace(".", "_")
    if randomize is None:
        return result

    chars = string.ascii_uppercase + string.digits
    return "{result}|{random_string}".format(
        result=result,
        random_string="".join(
            random.SystemRandom().choice(chars) for _ in range(randomize)
        ),
    )


def convert_path_parameter(match, types):
    name = match.group(1)
    swagger_type = types.get(name)
    converter = PATH_PARAMETER_CONVERTERS.get(swagger_type)
    return "<{}{}{}>".format(
        converter or "", ":" if converter else "", name.replace("-", "_")
    )


def flaskify_path(swagger_path, types=None):
    """
    Convert swagger path templates to flask path templates

    :type swagger_path: str
    :type types: dict
    :rtype: str

    >>> flaskify_path('/foo-bar/{my-param}')
    '/foo-bar/<my_param>'

    >>> flaskify_path('/foo/{someint}', {'someint': 'int'})
    '/foo/<int:someint>'
    """
    if types is None:
        types = {}
    convert_match = functools.partial(convert_path_parameter, types=types)
    return PATH_PARAMETER.sub(convert_match, swagger_path)


class FlaskJSONProvider(flask.json.provider.DefaultJSONProvider):
    """Custom JSONProvider which adds connexion defaults on top of Flask's"""

    @jsonifier.wrap_default
    def default(self, o):
        return super().default(o)


class NumberConverter(werkzeug.routing.BaseConverter):
    """Flask converter for OpenAPI number type"""

    regex = r"[+-]?[0-9]*(?:\.[0-9]*)?"

    def to_python(self, value):
        return float(value)


class IntegerConverter(werkzeug.routing.BaseConverter):
    """Flask converter for OpenAPI integer type"""

    regex = r"[+-]?[0-9]+"

    def to_python(self, value):
        return int(value)



import copy
import dataclasses
import enum
import logging
import pathlib
import typing as t
from dataclasses import dataclass, field
from functools import partial

from starlette.types import ASGIApp, Receive, Scope, Send

from connexion import utils
from connexion.handlers import ResolverErrorHandler
from connexion.jsonifier import Jsonifier
from connexion.lifecycle import ConnexionRequest, ConnexionResponse
from connexion.middleware.abstract import SpecMiddleware
from connexion.middleware.context import ContextMiddleware
from connexion.middleware.exceptions import ExceptionMiddleware
from connexion.middleware.lifespan import Lifespan, LifespanMiddleware
from connexion.middleware.request_validation import RequestValidationMiddleware
from connexion.middleware.response_validation import ResponseValidationMiddleware
from connexion.middleware.routing import RoutingMiddleware
from connexion.middleware.security import SecurityMiddleware
from connexion.middleware.server_error import ServerErrorMiddleware
from connexion.middleware.swagger_ui import SwaggerUIMiddleware
from connexion.options import SwaggerUIOptions
from connexion.resolver import Resolver
from connexion.spec import Specification
from connexion.types import MaybeAwaitable
from connexion.uri_parsing import AbstractURIParser
from connexion.utils import inspect_function_arguments

logger = logging.getLogger(__name__)


@dataclass
class _Options:
    """
    Connexion provides a lot of parameters for the user to configure the app / middleware of
    application.

    This class provides a central place to parse these parameters a mechanism to update them.
    Application level arguments can be provided when instantiating the application / middleware,
    after which they can be overwritten on an API level.

    The defaults should only be set in this class, and set to None in the signature of user facing
    methods. This is necessary for this class to be able to differentiate between missing and
    falsy arguments.
    """

    arguments: t.Optional[dict] = None
    auth_all_paths: t.Optional[bool] = False
    jsonifier: t.Optional[Jsonifier] = None
    pythonic_params: t.Optional[bool] = False
    resolver: t.Optional[t.Union[Resolver, t.Callable]] = None
    resolver_error: t.Optional[int] = None
    resolver_error_handler: t.Optional[t.Callable] = field(init=False)
    strict_validation: t.Optional[bool] = False
    swagger_ui_options: t.Optional[SwaggerUIOptions] = None
    uri_parser_class: t.Optional[AbstractURIParser] = None
    validate_responses: t.Optional[bool] = False
    validator_map: t.Optional[dict] = None
    security_map: t.Optional[dict] = None

    def __post_init__(self):
        self.resolver = (
            Resolver(self.resolver) if callable(self.resolver) else self.resolver
        )
        self.resolver_error_handler = self._resolver_error_handler_factory()

    def _resolver_error_handler_factory(
        self,
    ) -> t.Optional[t.Callable[[], ResolverErrorHandler]]:
        """Returns a factory to create a ResolverErrorHandler."""
        if self.resolver_error is not None:

            def resolver_error_handler(*args, **kwargs) -> ResolverErrorHandler:
                return ResolverErrorHandler(self.resolver_error, *args, **kwargs)

            return resolver_error_handler
        return None

    def replace(self, **changes) -> "_Options":
        """Update mechanism to overwrite the options. None values are discarded.

        :param changes: Arguments accepted by the __init__ method of this class.

        :return: An new _Options object with updated arguments.
        """
        changes = {key: value for key, value in changes.items() if value is not None}
        return dataclasses.replace(self, **changes)


class MiddlewarePosition(enum.Enum):
    """Positions to insert a middleware"""

    BEFORE_EXCEPTION = ExceptionMiddleware
    """Add before the :class:`ExceptionMiddleware`. This is useful if you want your changes to
    affect the way exceptions are handled, such as a custom error handler.

    Be mindful that security has not yet been applied at this stage.
    Additionally, the inserted middleware is positioned before the RoutingMiddleware, so you cannot
    leverage any routing information yet and should implement your middleware to work globally
    instead of on an operation level.

    Useful for middleware which should also be applied to error responses. Note that errors
    raised here will not be handled by the exception handlers and will always result in an
    internal server error response.

    :meta hide-value:
    """
    BEFORE_SWAGGER = SwaggerUIMiddleware
    """Add before the :class:`SwaggerUIMiddleware`. This is useful if you want your changes to
    affect the Swagger UI, such as a path altering middleware that should also alter the paths
    exposed by the Swagger UI

    Be mindful that security has not yet been applied at this stage.

    Since the inserted middleware is positioned before the RoutingMiddleware, you cannot leverage
    any routing information yet and should implement your middleware to work globally instead of on
    an operation level.

    :meta hide-value:
    """
    BEFORE_ROUTING = RoutingMiddleware
    """Add before the :class:`RoutingMiddleware`. This is useful if you want your changes to be
    applied before hitting the router, such as for path altering or CORS middleware.

    Be mindful that security has not yet been applied at this stage.

    Since the inserted middleware is positioned before the RoutingMiddleware, you cannot leverage
    any routing information yet and should implement your middleware to work globally instead of on
    an operation level.

    :meta hide-value:
    """
    BEFORE_SECURITY = SecurityMiddleware
    """Add before the :class:`SecurityMiddleware`. Insert middleware here that needs to be able to
    adapt incoming requests before security is applied.

    Be mindful that security has not yet been applied at this stage.

    Since the inserted middleware is positioned after the RoutingMiddleware, you can leverage
    routing information and implement the middleware to work on an individual operation level.

    :meta hide-value:
    """
    BEFORE_VALIDATION = RequestValidationMiddleware
    """Add before the :class:`RequestValidationMiddleware`. Insert middleware here that needs to be
    able to adapt incoming requests before they are validated.

    Since the inserted middleware is positioned after the RoutingMiddleware, you can leverage
    routing information and implement the middleware to work on an individual operation level.

    :meta hide-value:
    """
    BEFORE_CONTEXT = ContextMiddleware
    """Add before the :class:`ContextMiddleware`, near the end of the stack. This is the default
    location. The inserted middleware is only followed by the ContextMiddleware, which ensures any
    changes to the context are properly exposed to the application.

    Since the inserted middleware is positioned after the RoutingMiddleware, you can leverage
    routing information and implement the middleware to work on an individual operation level.

    Since the inserted middleware is positioned after the ResponseValidationMiddleware,
    it can intercept responses coming from the application and alter them before they are validated.

    :meta hide-value:
    """


class API:
    def __init__(self, specification, *, base_path, **kwargs) -> None:
        self.specification = specification
        self.base_path = base_path
        self.kwargs = kwargs


class ConnexionMiddleware:
    """The main Connexion middleware, which wraps a list of specialized middlewares around the
    provided application."""

    default_middlewares = [
        ServerErrorMiddleware,
        ExceptionMiddleware,
        SwaggerUIMiddleware,
        RoutingMiddleware,
        SecurityMiddleware,
        RequestValidationMiddleware,
        ResponseValidationMiddleware,
        LifespanMiddleware,
        ContextMiddleware,
    ]

    def __init__(
        self,
        app: ASGIApp,
        *,
        import_name: t.Optional[str] = None,
        lifespan: t.Optional[Lifespan] = None,
        middlewares: t.Optional[t.List[ASGIApp]] = None,
        specification_dir: t.Union[pathlib.Path, str] = "",
        arguments: t.Optional[dict] = None,
        auth_all_paths: t.Optional[bool] = None,
        jsonifier: t.Optional[Jsonifier] = None,
        pythonic_params: t.Optional[bool] = None,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = None,
        resolver_error: t.Optional[int] = None,
        strict_validation: t.Optional[bool] = None,
        swagger_ui_options: t.Optional[SwaggerUIOptions] = None,
        uri_parser_class: t.Optional[AbstractURIParser] = None,
        validate_responses: t.Optional[bool] = None,
        validator_map: t.Optional[dict] = None,
        security_map: t.Optional[dict] = None,
    ):
        """
        :param import_name: The name of the package or module that this object belongs to. If you
            are using a single module, __name__ is always the correct value. If you however are
            using a package, it’s usually recommended to hardcode the name of your package there.
        :param middlewares: The list of middlewares to wrap around the application. Defaults to
            :obj:`middleware.main.ConnexionmMiddleware.default_middlewares`
        :param specification_dir: The directory holding the specification(s). The provided path
            should either be absolute or relative to the root path of the application. Defaults to
            the root path.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate not paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: Instance of :class:`options.SwaggerUIOptions` with
            configuration options for the swagger ui.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`.
        :param security_map: A dictionary of security handlers to use. Defaults to
            :obj:`security.SECURITY_HANDLERS`.
        """
        import_name = import_name or str(pathlib.Path.cwd())
        self.root_path = utils.get_root_path(import_name)

        spec_dir = pathlib.Path(specification_dir)
        self.specification_dir = (
            spec_dir if spec_dir.is_absolute() else self.root_path / spec_dir
        )

        self.app = app
        self.lifespan = lifespan
        self.middlewares = (
            middlewares
            if middlewares is not None
            else copy.copy(self.default_middlewares)
        )
        self.middleware_stack: t.Optional[t.Iterable[ASGIApp]] = None
        self.apis: t.List[API] = []
        self.error_handlers: t.List[tuple] = []

        self.options = _Options(
            arguments=arguments,
            auth_all_paths=auth_all_paths,
            jsonifier=jsonifier,
            pythonic_params=pythonic_params,
            resolver=resolver,
            resolver_error=resolver_error,
            strict_validation=strict_validation,
            swagger_ui_options=swagger_ui_options,
            uri_parser_class=uri_parser_class,
            validate_responses=validate_responses,
            validator_map=validator_map,
            security_map=security_map,
        )

        self.extra_files: t.List[str] = []

    def add_middleware(
        self,
        middleware_class: t.Type[ASGIApp],
        *,
        position: MiddlewarePosition = MiddlewarePosition.BEFORE_CONTEXT,
        **options: t.Any,
    ) -> None:
        """Add a middleware to the stack on the specified position.

        :param middleware_class: Middleware class to add
        :param position: Position to add the middleware, one of the MiddlewarePosition Enum
        :param options: Options to pass to the middleware_class on initialization
        """
        if self.middleware_stack is not None:
            raise RuntimeError("Cannot add middleware after an application has started")

        for m, middleware in enumerate(self.middlewares):
            if isinstance(middleware, partial):
                middleware = middleware.func

            if middleware == position.value:
                self.middlewares.insert(
                    m, t.cast(ASGIApp, partial(middleware_class, **options))
                )
                break
        else:
            raise ValueError(
                f"Could not insert middleware at position {position.name}. "
                f"Please make sure you have a {position.value} in your stack."
            )

    def _build_middleware_stack(self) -> t.Tuple[ASGIApp, t.Iterable[ASGIApp]]:
        """Apply all middlewares to the provided app.

        :return: Tuple of the outer middleware wrapping the application and a list of the wrapped
            middlewares, including the wrapped application.
        """
        # Include the wrapped application in the returned list.
        app = self.app
        apps = [app]
        for middleware in reversed(self.middlewares):
            arguments, _ = inspect_function_arguments(middleware)
            if "lifespan" in arguments:
                app = middleware(app, lifespan=self.lifespan)  # type: ignore
            else:
                app = middleware(app)  # type: ignore
            apps.append(app)

        # We sort the APIs by base path so that the most specific APIs are registered first.
        # This is due to the way Starlette matches routes.
        self.apis = utils.sort_apis_by_basepath(self.apis)
        for app in apps:
            if isinstance(app, SpecMiddleware):
                for api in self.apis:
                    app.add_api(
                        api.specification,
                        base_path=api.base_path,
                        **api.kwargs,
                    )

            if isinstance(app, ExceptionMiddleware):
                for error_handler in self.error_handlers:
                    app.add_exception_handler(*error_handler)

        return app, list(reversed(apps))

    def add_api(
        self,
        specification: t.Union[pathlib.Path, str, dict],
        *,
        base_path: t.Optional[str] = None,
        name: t.Optional[str] = None,
        arguments: t.Optional[dict] = None,
        auth_all_paths: t.Optional[bool] = None,
        jsonifier: t.Optional[Jsonifier] = None,
        pythonic_params: t.Optional[bool] = None,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = None,
        resolver_error: t.Optional[int] = None,
        strict_validation: t.Optional[bool] = None,
        swagger_ui_options: t.Optional[SwaggerUIOptions] = None,
        uri_parser_class: t.Optional[AbstractURIParser] = None,
        validate_responses: t.Optional[bool] = None,
        validator_map: t.Optional[dict] = None,
        security_map: t.Optional[dict] = None,
        **kwargs,
    ) -> None:
        """
        Register een API represented by a single OpenAPI specification on this middleware.
        Multiple APIs can be registered on a single middleware.

        :param specification: OpenAPI specification. Can be provided either as dict, a path
            to file, or a URL.
        :param base_path: Base path to host the API. This overrides the basePath / servers in the
            specification.
        :param name: Name to register the API with. If no name is passed, the base_path is used
            as name instead.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate not paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: A dict with configuration options for the swagger ui. See
            :class:`options.SwaggerUIOptions`.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`
        :param security_map: A dictionary of security handlers to use. Defaults to
            :obj:`security.SECURITY_HANDLERS`
        :param kwargs: Additional keyword arguments to pass to the `add_api` method of the managed
            middlewares. This can be used to pass arguments to middlewares added beyond the default
            ones.
        """
        if self.middleware_stack is not None:
            raise RuntimeError("Cannot add api after an application has started")

        if isinstance(specification, str) and (
            specification.startswith("http://") or specification.startswith("https://")
        ):
            pass
        elif isinstance(specification, (pathlib.Path, str)):
            specification = t.cast(pathlib.Path, self.specification_dir / specification)

            # Add specification as file to watch for reloading
            if pathlib.Path.cwd() in specification.parents:
                self.extra_files.append(
                    str(specification.relative_to(pathlib.Path.cwd()))
                )

        specification = Specification.load(specification, arguments=arguments)

        options = self.options.replace(
            auth_all_paths=auth_all_paths,
            jsonifier=jsonifier,
            swagger_ui_options=swagger_ui_options,
            pythonic_params=pythonic_params,
            resolver=resolver,
            resolver_error=resolver_error,
            strict_validation=strict_validation,
            uri_parser_class=uri_parser_class,
            validate_responses=validate_responses,
            validator_map=validator_map,
            security_map=security_map,
        )

        api = API(
            specification, base_path=base_path, name=name, **options.__dict__, **kwargs
        )
        self.apis.append(api)

    def add_error_handler(
        self,
        code_or_exception: t.Union[int, t.Type[Exception]],
        function: t.Callable[
            [ConnexionRequest, Exception], MaybeAwaitable[ConnexionResponse]
        ],
    ) -> None:
        """
        Register a callable to handle application errors.

        :param code_or_exception: An exception class or the status code of HTTP exceptions to
            handle.
        :param function: Callable that will handle exception, may be async.
        """
        if self.middleware_stack is not None:
            raise RuntimeError(
                "Cannot add error handler after an application has started"
            )

        error_handler = (code_or_exception, function)
        self.error_handlers.append(error_handler)

    def run(self, import_string: t.Optional[str] = None, **kwargs):
        """Run the application using uvicorn.

        :param import_string: application as import string (eg. "main:app"). This is needed to run
                              using reload.
        :param kwargs: kwargs to pass to `uvicorn.run`.
        """
        try:
            import uvicorn
        except ImportError:
            raise RuntimeError(
                "uvicorn is not installed. Please install connexion using the uvicorn extra "
                "(connexion[uvicorn])"
            )

        logger.warning(
            f"`{self.__class__.__name__}.run` is optimized for development. "
            "For production, run using a dedicated ASGI server."
        )

        app: t.Union[str, ConnexionMiddleware]
        if import_string is not None:
            app = import_string
            kwargs.setdefault("reload", True)
            kwargs["reload_includes"] = self.extra_files + kwargs.get(
                "reload_includes", []
            )
        else:
            app = self

        uvicorn.run(app, **kwargs)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if self.middleware_stack is None:
            self.app, self.middleware_stack = self._build_middleware_stack()
        # Set so starlette router throws exceptions instead of returning error responses
        # This instance is also passed to any lifespan handler
        scope["app"] = self
        await self.app(scope, receive, send)



import collections
import copy
import logging

from jsonschema import Draft4Validator, ValidationError

from connexion.exceptions import BadRequestProblem, ExtraParameterProblem
from connexion.lifecycle import ConnexionRequest
from connexion.utils import boolean, is_null, is_nullable

logger = logging.getLogger("connexion.validators.parameter")

TYPE_MAP = {"integer": int, "number": float, "boolean": boolean, "object": dict}

try:
    draft4_format_checker = Draft4Validator.FORMAT_CHECKER  # type: ignore
except AttributeError:  # jsonschema < 4.5.0
    from jsonschema import draft4_format_checker


class ParameterValidator:
    def __init__(
        self,
        parameters,
        uri_parser,
        strict_validation=False,
        security_query_params=None,
    ):
        """
        :param parameters: List of request parameter dictionaries
        :param uri_parser: class to use for uri parsing
        :param strict_validation: Flag indicating if parameters not in spec are allowed
        :param security_query_params: List of query parameter names used for security.
            These parameters will be ignored when checking for extra parameters in case of
            strict validation.
        """
        self.parameters = collections.defaultdict(list)
        for p in parameters:
            self.parameters[p["in"]].append(p)

        self.uri_parser = uri_parser
        self.strict_validation = strict_validation
        self.security_query_params = set(security_query_params or [])

    @staticmethod
    def validate_parameter(parameter_type, value, param, param_name=None):
        if is_nullable(param) and is_null(value):
            return

        elif value is not None:
            param = copy.deepcopy(param)
            param = param.get("schema", param)
            try:
                Draft4Validator(param, format_checker=draft4_format_checker).validate(
                    value
                )
            except ValidationError as exception:
                return str(exception)

        elif param.get("required"):
            return "Missing {parameter_type} parameter '{param[name]}'".format(
                **locals()
            )

    @staticmethod
    def validate_parameter_list(request_params, spec_params):
        request_params = set(request_params)
        spec_params = set(spec_params)

        return request_params.difference(spec_params)

    def validate_query_parameter_list(self, request, security_params=None):
        request_params = request.query_params.keys()
        spec_params = [x["name"] for x in self.parameters.get("query", [])]
        spec_params.extend(security_params or [])
        return self.validate_parameter_list(request_params, spec_params)

    def validate_query_parameter(self, param, request):
        """
        Validate a single query parameter (request.args in Flask)

        :type param: dict
        :rtype: str
        """
        val = request.query_params.get(param["name"])
        return self.validate_parameter("query", val, param)

    def validate_path_parameter(self, param, request):
        val = request.path_params.get(param["name"].replace("-", "_"))
        return self.validate_parameter("path", val, param)

    def validate_header_parameter(self, param, request):
        val = request.headers.get(param["name"])
        return self.validate_parameter("header", val, param)

    def validate_cookie_parameter(self, param, request):
        val = request.cookies.get(param["name"])
        return self.validate_parameter("cookie", val, param)

    def validate(self, scope):
        logger.debug("%s validating parameters...", scope.get("path"))

        request = ConnexionRequest(scope, uri_parser=self.uri_parser)
        self.validate_request(request)

    def validate_request(self, request):
        if self.strict_validation:
            query_errors = self.validate_query_parameter_list(
                request, security_params=self.security_query_params
            )

            if query_errors:
                raise ExtraParameterProblem(
                    param_type="query", extra_params=query_errors
                )

        for param in self.parameters.get("query", []):
            error = self.validate_query_parameter(param, request)
            if error:
                raise BadRequestProblem(detail=error)

        for param in self.parameters.get("path", []):
            error = self.validate_path_parameter(param, request)
            if error:
                raise BadRequestProblem(detail=error)

        for param in self.parameters.get("header", []):
            error = self.validate_header_parameter(param, request)
            if error:
                raise BadRequestProblem(detail=error)

        for param in self.parameters.get("cookie", []):
            error = self.validate_cookie_parameter(param, request)
            if error:
                raise BadRequestProblem(detail=error)



"""
This module defines a command-line interface (CLI) that runs an OpenAPI specification to be a
starting point for developing your API with Connexion.
"""

import argparse
import importlib.metadata
import logging
import os
import sys
import typing as t

import connexion
from connexion.apps import AbstractApp
from connexion.mock import MockResolver
from connexion.options import SwaggerUIOptions

logger = logging.getLogger(__name__)

FLASK_APP = "flask"
ASYNC_APP = "async"
AVAILABLE_APPS = {
    FLASK_APP: "connexion.apps.flask.FlaskApp",
    ASYNC_APP: "connexion.apps.asynchronous.AsyncApp",
}


def run(app: AbstractApp, args: argparse.Namespace):
    app.run("connexion.cli:create_app", port=args.port, host=args.host, factory=True)


parser = argparse.ArgumentParser()

parser.add_argument(
    "--version",
    action="version",
    version=f"Connexion {importlib.metadata.version('connexion')}",
)

subparsers = parser.add_subparsers()
run_parser = subparsers.add_parser("run")
run_parser.set_defaults(func=run)

run_parser.add_argument("spec_file", help="Path to OpenAPI specification.")
run_parser.add_argument(
    "base_module_path", nargs="?", help="Root directory of handler code."
)
run_parser.add_argument(
    "-p", "--port", default=5000, type=int, help="Port to listen on."
)
run_parser.add_argument(
    "-H", "--host", default="127.0.0.1", type=str, help="Host interface to bind on."
)
run_parser.add_argument(
    "--stub",
    action="store_true",
    help="Returns status code 501, and `Not Implemented Yet` payload, for the endpoints which "
    "handlers are not found.",
)
run_parser.add_argument(
    "--mock",
    choices=["all", "notimplemented"],
    help="Returns example data for all endpoints or for which handlers are not found.",
)
run_parser.add_argument(
    "--swagger-ui-path",
    help="Personalize what URL path the API console UI will be mounted.",
    default="/ui",
)
run_parser.add_argument(
    "--swagger-ui-template-dir",
    help="Path to a customized API console UI dashboard.",
)
run_parser.add_argument(
    "--auth-all-paths",
    help="Enable authentication to paths not defined in the spec.",
    action="store_true",
)
run_parser.add_argument(
    "--validate-responses",
    help="Enable validation of response values from operation handlers.",
    action="store_true",
)
run_parser.add_argument(
    "--strict-validation",
    help="Enable strict validation of request payloads.",
    action="store_true",
)
run_parser.add_argument(
    "-v",
    "--verbose",
    help="Show verbose information.",
    action="count",
    default=0,
)
run_parser.add_argument("--base-path", help="Override the basePath in the API spec.")
run_parser.add_argument(
    "--app-framework",
    "-f",
    choices=list(AVAILABLE_APPS),
    default=ASYNC_APP,
    help="The app framework used to run the server",
)


def create_app(args: t.Optional[argparse.Namespace] = None) -> AbstractApp:
    """Runs a server compliant with a OpenAPI/Swagger Specification file."""
    if args is None:
        args = parser.parse_args()

    if args.verbose == 1:
        logging_level = logging.INFO
    elif args.verbose >= 2:
        logging_level = logging.DEBUG
    else:
        logging_level = logging.WARN

    logging.basicConfig(level=logging_level)

    if args.spec_file.startswith("http") or args.spec_file.startswith("https"):
        spec_file_full_path = args.spec_file
    else:
        spec_file_full_path = os.path.abspath(args.spec_file)
        py_module_path = args.base_module_path or os.path.dirname(spec_file_full_path)
        sys.path.insert(1, os.path.abspath(py_module_path))
        logger.debug(f"Added {py_module_path} to system path.")

    resolver_error = None
    if args.stub:
        resolver_error = 501

    api_extra_args = {}
    if args.mock:
        resolver = MockResolver(mock_all=args.mock == "all")
        api_extra_args["resolver"] = resolver

    app_cls = connexion.utils.get_function_from_name(AVAILABLE_APPS[args.app_framework])

    swagger_ui_options = SwaggerUIOptions(
        swagger_ui_path=args.swagger_ui_path,
        swagger_ui_template_dir=args.swagger_ui_template_dir,
    )

    app = app_cls(
        __name__,
        auth_all_paths=args.auth_all_paths,
        swagger_ui_options=swagger_ui_options,
    )

    app.add_api(
        spec_file_full_path,
        base_path=args.base_path,
        resolver_error=resolver_error,
        validate_responses=args.validate_responses,
        strict_validation=args.strict_validation,
        **api_extra_args,
    )

    return app


def main(argv: t.Optional[t.List[str]] = None) -> None:
    if argv is None:
        argv = sys.argv[1:]
    if not argv:
        argv = ["--help"]

    args = parser.parse_args(argv)
    app = create_app(args)
    args.func(app, args)



import json
import logging
import typing as t

import jsonschema
from jsonschema import Draft4Validator, ValidationError
from starlette.types import Scope

from connexion.exceptions import BadRequestProblem, NonConformingResponseBody
from connexion.json_schema import (
    Draft4RequestValidator,
    Draft4ResponseValidator,
    format_error_with_path,
)
from connexion.validators import (
    AbstractRequestBodyValidator,
    AbstractResponseBodyValidator,
)

logger = logging.getLogger(__name__)


class JSONRequestBodyValidator(AbstractRequestBodyValidator):
    """Request body validator for json content types."""

    def __init__(
        self,
        *,
        schema: dict,
        required=False,
        nullable=False,
        encoding: str,
        strict_validation: bool,
        **kwargs,
    ) -> None:
        super().__init__(
            schema=schema,
            required=required,
            nullable=nullable,
            encoding=encoding,
            strict_validation=strict_validation,
        )

    @property
    def _validator(self):
        return Draft4RequestValidator(
            self._schema, format_checker=Draft4Validator.FORMAT_CHECKER
        )

    async def _parse(
        self, stream: t.AsyncGenerator[bytes, None], scope: Scope
    ) -> t.Any:
        bytes_body = b"".join([message async for message in stream])
        body = bytes_body.decode(self._encoding)

        if not body:
            return None

        try:
            return json.loads(body)
        except json.decoder.JSONDecodeError as e:
            raise BadRequestProblem(detail=str(e))

    def _validate(self, body: t.Any) -> t.Optional[dict]:
        if not self._nullable and body is None:
            raise BadRequestProblem("Request body must not be empty")
        try:
            return self._validator.validate(body)
        except ValidationError as exception:
            error_path_msg = format_error_with_path(exception=exception)
            logger.info(
                f"Validation error: {exception.message}{error_path_msg}",
                extra={"validator": "body"},
            )
            raise BadRequestProblem(detail=f"{exception.message}{error_path_msg}")


class DefaultsJSONRequestBodyValidator(JSONRequestBodyValidator):
    """Request body validator for json content types which fills in default values. This Validator
    intercepts the body, makes changes to it, and replays it for the next ASGI application.
    """

    MUTABLE_VALIDATION = True
    """This validator might mutate to the body."""

    @property
    def _validator(self):
        validator_cls = self.extend_with_set_default(Draft4RequestValidator)
        return validator_cls(
            self._schema, format_checker=Draft4Validator.FORMAT_CHECKER
        )

    # via https://python-jsonschema.readthedocs.io/
    @staticmethod
    def extend_with_set_default(validator_class):
        validate_properties = validator_class.VALIDATORS["properties"]

        def set_defaults(validator, properties, instance, schema):
            for property, subschema in properties.items():
                if "default" in subschema:
                    instance.setdefault(property, subschema["default"])

            yield from validate_properties(validator, properties, instance, schema)

        return jsonschema.validators.extend(
            validator_class, {"properties": set_defaults}
        )


class JSONResponseBodyValidator(AbstractResponseBodyValidator):
    """Response body validator for json content types."""

    @property
    def validator(self) -> Draft4Validator:
        return Draft4ResponseValidator(
            self._schema, format_checker=Draft4Validator.FORMAT_CHECKER
        )

    def _parse(self, stream: t.Generator[bytes, None, None]) -> t.Any:
        body = b"".join(stream).decode(self._encoding)

        if not body:
            return None

        try:
            return json.loads(body)
        except json.decoder.JSONDecodeError as e:
            raise NonConformingResponseBody(str(e))

    def _validate(self, body: dict):
        try:
            self.validator.validate(body)
        except ValidationError as exception:
            error_path_msg = format_error_with_path(exception=exception)
            logger.warning(
                f"Validation error: {exception.message}{error_path_msg}",
                extra={"validator": "body"},
            )
            raise NonConformingResponseBody(
                detail=f"Response body does not conform to specification. {exception.message}{error_path_msg}"
            )


class TextResponseBodyValidator(JSONResponseBodyValidator):
    def _parse(self, stream: t.Generator[bytes, None, None]) -> str:  # type: ignore
        body = b"".join(stream).decode(self._encoding)

        try:
            return json.loads(body)
        except json.decoder.JSONDecodeError:
            return body



"""
This module defines an AbstractApp, which defines a standardized user interface for a Connexion
application.
"""
import abc
import pathlib
import typing as t

from starlette.testclient import TestClient
from starlette.types import ASGIApp, Receive, Scope, Send

from connexion.jsonifier import Jsonifier
from connexion.lifecycle import ConnexionRequest, ConnexionResponse
from connexion.middleware import ConnexionMiddleware, MiddlewarePosition, SpecMiddleware
from connexion.middleware.lifespan import Lifespan
from connexion.options import SwaggerUIOptions
from connexion.resolver import Resolver
from connexion.types import MaybeAwaitable
from connexion.uri_parsing import AbstractURIParser


class AbstractApp:
    """
    Abstract class for a Connexion Application. A Connexion Application provides an interface for a
    framework application wrapped by Connexion Middleware. Since its main function is to provide an
    interface, it delegates most of the work to the middleware and framework application.
    """

    _middleware_app: SpecMiddleware
    """
    The application wrapped by the ConnexionMiddleware, which in its turn wraps the framework
    application.
    """

    def __init__(
        self,
        import_name: str,
        *,
        lifespan: t.Optional[Lifespan] = None,
        middlewares: t.Optional[list] = None,
        specification_dir: t.Union[pathlib.Path, str] = "",
        arguments: t.Optional[dict] = None,
        auth_all_paths: t.Optional[bool] = None,
        jsonifier: t.Optional[Jsonifier] = None,
        pythonic_params: t.Optional[bool] = None,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = None,
        resolver_error: t.Optional[int] = None,
        strict_validation: t.Optional[bool] = None,
        swagger_ui_options: t.Optional[SwaggerUIOptions] = None,
        uri_parser_class: t.Optional[AbstractURIParser] = None,
        validate_responses: t.Optional[bool] = None,
        validator_map: t.Optional[dict] = None,
        security_map: t.Optional[dict] = None,
    ) -> None:
        """
        :param import_name: The name of the package or module that this object belongs to. If you
            are using a single module, __name__ is always the correct value. If you however are
            using a package, it’s usually recommended to hardcode the name of your package there.
        :param lifespan: A lifespan context function, which can be used to perform startup and
        :param middlewares: The list of middlewares to wrap around the application. Defaults to
            :obj:`middleware.main.ConnexionMiddleware.default_middlewares`
        :param specification_dir: The directory holding the specification(s). The provided path
            should either be absolute or relative to the root path of the application. Defaults to
            the root path.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate not paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: Instance of :class:`options.SwaggerUIOptions` with
            configuration options for the swagger ui.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`.
        :param security_map: A dictionary of security handlers to use. Defaults to
            :obj:`security.SECURITY_HANDLERS`
        """
        self.middleware = ConnexionMiddleware(
            self._middleware_app,
            import_name=import_name,
            lifespan=lifespan,
            middlewares=middlewares,
            specification_dir=specification_dir,
            arguments=arguments,
            auth_all_paths=auth_all_paths,
            jsonifier=jsonifier,
            swagger_ui_options=swagger_ui_options,
            pythonic_params=pythonic_params,
            resolver=resolver,
            resolver_error=resolver_error,
            strict_validation=strict_validation,
            uri_parser_class=uri_parser_class,
            validate_responses=validate_responses,
            validator_map=validator_map,
            security_map=security_map,
        )

    def add_middleware(
        self,
        middleware_class: t.Type[ASGIApp],
        position: MiddlewarePosition = MiddlewarePosition.BEFORE_CONTEXT,
        **options: t.Any,
    ) -> None:
        """Add a middleware to the stack on the specified position.

        :param middleware_class: Middleware class to add
        :param position: Position to add the middleware, one of the MiddlewarePosition Enum
        :param options: Options to pass to the middleware_class on initialization
        """
        self.middleware.add_middleware(middleware_class, position=position, **options)

    def add_api(
        self,
        specification: t.Union[pathlib.Path, str, dict],
        *,
        base_path: t.Optional[str] = None,
        name: t.Optional[str] = None,
        arguments: t.Optional[dict] = None,
        auth_all_paths: t.Optional[bool] = None,
        jsonifier: t.Optional[Jsonifier] = None,
        pythonic_params: t.Optional[bool] = None,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = None,
        resolver_error: t.Optional[int] = None,
        strict_validation: t.Optional[bool] = None,
        swagger_ui_options: t.Optional[SwaggerUIOptions] = None,
        uri_parser_class: t.Optional[AbstractURIParser] = None,
        validate_responses: t.Optional[bool] = None,
        validator_map: t.Optional[dict] = None,
        security_map: t.Optional[dict] = None,
        **kwargs,
    ) -> t.Any:
        """
        Register an API represented by a single OpenAPI specification on this application.
        Multiple APIs can be registered on a single application.

        :param specification: OpenAPI specification. Can be provided either as dict, a path
            to file, or a URL.
        :param base_path: Base path to host the API. This overrides the basePath / servers in the
            specification.
        :param name: Name to register the API with. If no name is passed, the base_path is used
            as name instead.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate not paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: A :class:`options.SwaggerUIOptions` instance with configuration
            options for the swagger ui.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`
        :param security_map: A dictionary of security handlers to use. Defaults to
            :obj:`security.SECURITY_HANDLERS`
        :param kwargs: Additional keyword arguments to pass to the `add_api` method of the managed
            middlewares. This can be used to pass arguments to middlewares added beyond the default
            ones.

        :return: The Api registered on the middleware application wrapping the framework.
        """
        return self.middleware.add_api(
            specification,
            base_path=base_path,
            name=name,
            arguments=arguments,
            auth_all_paths=auth_all_paths,
            jsonifier=jsonifier,
            pythonic_params=pythonic_params,
            resolver=resolver,
            resolver_error=resolver_error,
            strict_validation=strict_validation,
            swagger_ui_options=swagger_ui_options,
            uri_parser_class=uri_parser_class,
            validate_responses=validate_responses,
            validator_map=validator_map,
            security_map=security_map,
            **kwargs,
        )

    def add_url_rule(
        self,
        rule,
        endpoint: t.Optional[str] = None,
        view_func: t.Optional[t.Callable] = None,
        **options,
    ):
        """
        Connects a URL rule.  Works exactly like the `route` decorator.

        Basically this example::

            @app.route('/')
            def index():
                pass

        Is equivalent to the following::

            def index():
                pass
            app.add_url_rule('/', 'index', index)

        Internally`route` invokes `add_url_rule` so if you want to customize the behavior via
        subclassing you only need to change this method.

        :param rule: the URL rule as string.
        :param endpoint: the name of the endpoint for the registered URL rule, which is used for
            reverse lookup. Flask defaults to the name of the view function.
        :param view_func: the function to call when serving a request to the provided endpoint.
        :param options: the options to be forwarded to the underlying ``werkzeug.routing.Rule``
            object.  A change to Werkzeug is handling of method options. methods is a list of
            methods this rule should be limited to (`GET`, `POST` etc.).  By default a rule just
            listens for `GET` (and implicitly `HEAD`).
        """

    def route(self, rule: str, **options):
        """
        A decorator that is used to register a view function for a
        given URL rule.  This does the same thing as `add_url_rule`
        but is intended for decorator usage::

            @app.route('/')
            def index():
                return 'Hello World'

        :param rule: the URL rule as string
        :param options: the options to be forwarded to the underlying ``werkzeug.routing.Rule``
                        object. A change to Werkzeug is handling of method options. methods is a
                        list of methods this rule should be limited to (`GET`, `POST` etc.).
                        By default a rule just listens for `GET` (and implicitly `HEAD`).
        """

        def decorator(func: t.Callable) -> t.Callable:
            self.add_url_rule(rule, view_func=func, **options)
            return func

        return decorator

    @abc.abstractmethod
    def add_error_handler(
        self,
        code_or_exception: t.Union[int, t.Type[Exception]],
        function: t.Callable[
            [ConnexionRequest, Exception], MaybeAwaitable[ConnexionResponse]
        ],
    ) -> None:
        """
        Register a callable to handle application errors.

        :param code_or_exception: An exception class or the status code of HTTP exceptions to
            handle.
        :param function: Callable that will handle exception, may be async.
        """

    def test_client(self, **kwargs):
        """Creates a test client for this application. The keywords arguments passed in are
        passed to the ``StarletteClient``."""
        return TestClient(self, **kwargs)

    def run(self, import_string: t.Optional[str] = None, **kwargs):
        """Run the application using uvicorn.

        :param import_string: application as import string (eg. "main:app"). This is needed to run
                              using reload.
        :param kwargs: kwargs to pass to `uvicorn.run`.
        """
        self.middleware.run(import_string, **kwargs)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        return await self.middleware(scope, receive, send)



"""The ContextMiddleware creates a global context based the scope. It should be last in the
middleware stack, so it exposes the scope passed to the application"""
from starlette.types import ASGIApp, Receive, Scope, Send

from connexion.context import _context, _operation, _receive, _scope
from connexion.middleware.abstract import RoutedAPI, RoutedMiddleware
from connexion.operations import AbstractOperation


class ContextOperation:
    def __init__(
        self,
        next_app: ASGIApp,
        *,
        operation: AbstractOperation,
    ) -> None:
        self.next_app = next_app
        self.operation = operation

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        _context.set(scope.get("extensions", {}).get("connexion_context", {}))
        _operation.set(self.operation)
        _receive.set(receive)
        _scope.set(scope)
        await self.next_app(scope, receive, send)


class ContextAPI(RoutedAPI[ContextOperation]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.add_paths()

    def make_operation(self, operation: AbstractOperation) -> ContextOperation:
        return ContextOperation(self.next_app, operation=operation)


class ContextMiddleware(RoutedMiddleware[ContextAPI]):
    """Middleware to expose operation specific context to application."""

    api_cls = ContextAPI



"""
This module defines a Validator interface with base functionality that can be subclassed
for custom validators provided to the RequestValidationMiddleware.
"""
import copy
import json
import typing as t

from starlette.datastructures import Headers, MutableHeaders
from starlette.types import Receive, Scope, Send

from connexion.exceptions import BadRequestProblem


class AbstractRequestBodyValidator:
    """
    Validator interface with base functionality that can be subclassed for custom validators.

    .. note: Validators load the whole body into memory, which can be a problem for large payloads.
    """

    MUTABLE_VALIDATION = False
    """
    Whether mutations to the body during validation should be transmitted via the receive channel.
    Note that this does not apply to the substitution of a missing body with the default body, which always
    updates the receive channel.
    """
    MAX_MESSAGE_LENGTH = 256000
    """Maximum message length that will be sent via the receive channel for mutated bodies."""

    def __init__(
        self,
        *,
        schema: dict,
        required: bool = False,
        nullable: bool = False,
        encoding: str,
        strict_validation: bool,
        **kwargs,
    ):
        """
        :param schema: Schema of operation to validate
        :param required: Whether RequestBody is required
        :param nullable: Whether RequestBody is nullable
        :param encoding: Encoding of body (passed via Content-Type header)
        :param kwargs: Additional arguments for subclasses
        :param strict_validation: Whether to allow parameters not defined in the spec
        """
        self._schema = schema
        self._nullable = nullable
        self._required = required
        self._encoding = encoding
        self._strict_validation = strict_validation

    async def _parse(
        self, stream: t.AsyncGenerator[bytes, None], scope: Scope
    ) -> t.Any:
        """Parse the incoming stream."""

    def _validate(self, body: t.Any) -> t.Optional[dict]:
        """
        Validate the parsed body.

        :raises: :class:`connexion.exceptions.BadRequestProblem`
        """

    def _insert_body(
        self, receive: Receive, *, body: t.Any, scope: Scope
    ) -> t.Tuple[Receive, Scope]:
        """
        Insert messages transmitting the body at the start of the `receive` channel.

        This method updates the provided `scope` in place with the right `Content-Length` header.
        """
        if body is None:
            return receive, scope

        bytes_body = json.dumps(body).encode(self._encoding)

        # Update the content-length header
        new_scope = scope.copy()
        new_scope["headers"] = copy.deepcopy(scope["headers"])
        headers = MutableHeaders(scope=new_scope)
        headers["content-length"] = str(len(bytes_body))

        # Wrap in new receive channel
        messages = (
            {
                "type": "http.request",
                "body": bytes_body[i : i + self.MAX_MESSAGE_LENGTH],
                "more_body": i + self.MAX_MESSAGE_LENGTH < len(bytes_body),
            }
            for i in range(0, len(bytes_body), self.MAX_MESSAGE_LENGTH)
        )

        receive = self._insert_messages(receive, messages=messages)

        return receive, new_scope

    @staticmethod
    def _insert_messages(
        receive: Receive, *, messages: t.Iterable[t.MutableMapping[str, t.Any]]
    ) -> Receive:
        """Insert messages at the start of the `receive` channel."""
        # Ensure that messages is an iterator so each message is replayed once.
        message_iterator = iter(messages)

        async def receive_() -> t.MutableMapping[str, t.Any]:
            try:
                return next(message_iterator)
            except StopIteration:
                return await receive()

        return receive_

    async def wrap_receive(
        self, receive: Receive, *, scope: Scope
    ) -> t.Tuple[Receive, Scope]:
        """
        Wrap the provided `receive` channel with request body validation.

        This method updates the provided `scope` in place with the right `Content-Length` header.
        """
        # Handle missing bodies
        headers = Headers(scope=scope)
        if not int(headers.get("content-length", 0)):
            body = self._schema.get("default")
            if body is None and self._required:
                raise BadRequestProblem("RequestBody is required")
            # The default body is encoded as a `receive` channel to mimic an incoming body
            receive, scope = self._insert_body(receive, body=body, scope=scope)

        # The receive channel is converted to a stream for convenient access
        messages = []

        async def stream() -> t.AsyncGenerator[bytes, None]:
            more_body = True
            while more_body:
                message = await receive()
                messages.append(message)
                more_body = message.get("more_body", False)
                yield message.get("body", b"")
            yield b""

        # The body is parsed and validated
        body = await self._parse(stream(), scope=scope)
        if not (body is None and self._nullable):
            self._validate(body)

        # If MUTABLE_VALIDATION is enabled, include any changes made during validation in the messages to send
        if self.MUTABLE_VALIDATION:
            # Include changes made during validation
            receive, scope = self._insert_body(receive, body=body, scope=scope)
        else:
            # Serialize original messages
            receive = self._insert_messages(receive, messages=messages)

        return receive, scope


class AbstractResponseBodyValidator:
    """
    Validator interface with base functionality that can be subclassed for custom validators.

    .. note: Validators load the whole body into memory, which can be a problem for large payloads.
    """

    def __init__(
        self,
        scope: Scope,
        *,
        schema: dict,
        nullable: bool = False,
        encoding: str,
    ) -> None:
        self._scope = scope
        self._schema = schema
        self._nullable = nullable
        self._encoding = encoding

    def _parse(self, stream: t.Generator[bytes, None, None]) -> t.Any:
        """Parse the incoming stream."""

    def _validate(self, body: t.Any) -> t.Optional[dict]:
        """
        Validate the body.

        :raises: :class:`connexion.exceptions.NonConformingResponse`
        """

    def wrap_send(self, send: Send) -> Send:
        """Wrap the provided send channel with response body validation"""

        messages = []

        async def send_(message: t.MutableMapping[str, t.Any]) -> None:
            messages.append(message)

            if message["type"] == "http.response.start" or message.get(
                "more_body", False
            ):
                return

            stream = (message.get("body", b"") for message in messages)
            body = self._parse(stream)

            if not (body is None and self._nullable):
                self._validate(body)

            while messages:
                await send(messages.pop(0))

        return send_



import logging
import typing as t
from collections import defaultdict

from starlette.types import ASGIApp, Receive, Scope, Send

from connexion.exceptions import ProblemException
from connexion.lifecycle import ConnexionRequest
from connexion.middleware.abstract import RoutedAPI, RoutedMiddleware
from connexion.operations import AbstractOperation
from connexion.security import SecurityHandlerFactory
from connexion.spec import Specification

logger = logging.getLogger("connexion.middleware.security")


class SecurityOperation:
    def __init__(
        self,
        next_app: ASGIApp,
        *,
        security_handler_factory: SecurityHandlerFactory,
        security: list,
        security_schemes: dict,
    ):
        self.next_app = next_app
        self.security_handler_factory = security_handler_factory
        self.security = security
        self.security_schemes = security_schemes
        self.verification_fn = self._get_verification_fn()

    @classmethod
    def from_operation(
        cls,
        operation: t.Union[AbstractOperation, Specification],
        *,
        next_app: ASGIApp,
        security_handler_factory: SecurityHandlerFactory,
    ) -> "SecurityOperation":
        """Create a SecurityOperation from an Operation of Specification instance

        :param operation: The operation can be both an Operation or Specification instance here
            since security is defined at both levels in the OpenAPI spec. Creating a
            SecurityOperation based on a Specification can be used to create a SecurityOperation
            for routes not explicitly defined in the specification.
        :param next_app: The next ASGI app to call.
        :param security_handler_factory: The factory to be used to generate security handlers for
            the different security schemes.
        """
        return cls(
            next_app=next_app,
            security_handler_factory=security_handler_factory,
            security=operation.security,
            security_schemes=operation.security_schemes,
        )

    def _get_verification_fn(self):
        logger.debug("... Security: %s", self.security, extra=vars(self))
        if not self.security:
            return self.security_handler_factory.security_passthrough

        auth_funcs = []
        for security_req in self.security:
            if not security_req:
                auth_funcs.append(self.security_handler_factory.verify_none)
                continue

            sec_req_funcs = {}
            oauth = False
            for scheme_name, required_scopes in security_req.items():
                security_scheme = self.security_schemes[scheme_name]

                if security_scheme["type"] == "oauth2":
                    if oauth:
                        logger.warning(
                            "... multiple OAuth2 security schemes in AND fashion not supported",
                            extra=vars(self),
                        )
                        break
                    oauth = True

                sec_req_func = self.security_handler_factory.parse_security_scheme(
                    security_scheme, required_scopes
                )
                if sec_req_func is None:
                    break

                sec_req_funcs[scheme_name] = sec_req_func

            else:
                # No break encountered: no missing funcs
                if len(sec_req_funcs) == 1:
                    (func,) = sec_req_funcs.values()
                    auth_funcs.append(func)
                else:
                    auth_funcs.append(
                        self.security_handler_factory.verify_multiple_schemes(
                            sec_req_funcs
                        )
                    )

        return self.security_handler_factory.verify_security(auth_funcs)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if not self.security:
            await self.next_app(scope, receive, send)
            return

        request = ConnexionRequest(scope)
        await self.verification_fn(request)
        await self.next_app(scope, receive, send)


class SecurityAPI(RoutedAPI[SecurityOperation]):
    def __init__(
        self,
        *args,
        auth_all_paths: bool = False,
        security_map: t.Optional[dict] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.security_handler_factory = SecurityHandlerFactory(security_map)

        if auth_all_paths:
            self.add_auth_on_not_found()
        else:
            self.operations: t.MutableMapping[t.Optional[str], SecurityOperation] = {}

        self.add_paths()

    def add_auth_on_not_found(self) -> None:
        """Register a default SecurityOperation for routes that are not found."""
        default_operation = self.make_operation(self.specification)
        self.operations = defaultdict(lambda: default_operation)

    def make_operation(
        self, operation: t.Union[AbstractOperation, Specification]
    ) -> SecurityOperation:
        """Create a SecurityOperation from an Operation of Specification instance

        :param operation: The operation can be both an Operation or Specification instance here
            since security is defined at both levels in the OpenAPI spec. Creating a
            SecurityOperation based on a Specification can be used to create a SecurityOperation
            for routes not explicitly defined in the specification.
        """
        return SecurityOperation.from_operation(
            operation,
            next_app=self.next_app,
            security_handler_factory=self.security_handler_factory,
        )


class SecurityMiddleware(RoutedMiddleware[SecurityAPI]):
    """Middleware to check if operation is accessible on scope."""

    api_cls = SecurityAPI


class MissingSecurityOperation(ProblemException):
    pass



"""
This module defines error handlers, operations that produce proper response problems.
"""

import logging

from .exceptions import ResolverProblem

logger = logging.getLogger("connexion.handlers")

RESOLVER_ERROR_ENDPOINT_RANDOM_DIGITS = 6


class ResolverErrorHandler:
    """
    Handler for responding to ResolverError.
    """

    def __init__(self, status_code, exception):
        self.status_code = status_code
        self.exception = exception

    @property
    def function(self):
        return self.handle

    def handle(self, *args, **kwargs):
        raise ResolverProblem(
            detail=self.exception.args[0],
            status=self.status_code,
        )

    @property
    def operation_id(self):
        return "noop"

    @property
    def randomize_endpoint(self):
        return RESOLVER_ERROR_ENDPOINT_RANDOM_DIGITS

    def get_path_parameter_types(self):
        return {}

    @property
    def uri_parser_class(self):
        return "dummy"

    @property
    def api(self):
        return "dummy"

    def get_mimetype(self):
        return "dummy"

    async def __call__(self, *args, **kwargs):
        raise ResolverProblem(
            detail=self.exception.args[0],
            status=self.status_code,
        )



import abc
import typing as t

from typing_extensions import Protocol


class Framework(Protocol):
    @staticmethod
    @abc.abstractmethod
    def is_framework_response(response: t.Any) -> bool:
        """Return True if provided response is a framework type"""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def connexion_to_framework_response(cls, response):
        """Cast ConnexionResponse to framework response class"""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def build_response(
        cls,
        data: t.Any,
        *,
        content_type: t.Optional[str] = None,
        headers: t.Optional[dict] = None,
        status_code: int = None
    ):
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_request(*args, **kwargs):
        """Return a framework request from the context."""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_body(request):
        """Get body from a framework request"""
        raise NotImplementedError



"""
This module defines URIParsers which parse query and path parameters according to OpenAPI
serialization rules.
"""

import abc
import json
import logging
import re

from connexion.exceptions import TypeValidationError
from connexion.utils import all_json, coerce_type, deep_merge

logger = logging.getLogger("connexion.decorators.uri_parsing")

QUERY_STRING_DELIMITERS = {
    "spaceDelimited": " ",
    "pipeDelimited": "|",
    "simple": ",",
    "form": ",",
}


class AbstractURIParser(metaclass=abc.ABCMeta):
    parsable_parameters = ["query", "path"]

    def __init__(self, param_defns, body_defn):
        """
        a URI parser is initialized with parameter definitions.
        When called with a request object, it handles array types in the URI
        both in the path and query according to the spec.
        Some examples include:
        - https://mysite.fake/in/path/1,2,3/            # path parameters
        - https://mysite.fake/?in_query=a,b,c           # simple query params
        - https://mysite.fake/?in_query=a|b|c           # various separators
        - https://mysite.fake/?in_query=a&in_query=b,c  # complex query params
        """
        self._param_defns = {
            p["name"]: p for p in param_defns if p["in"] in self.parsable_parameters
        }
        self._body_schema = body_defn.get("schema", {})
        self._body_encoding = body_defn.get("encoding", {})

    @property
    @abc.abstractmethod
    def param_defns(self):
        """
        returns the parameter definitions by name
        """

    @property
    @abc.abstractmethod
    def param_schemas(self):
        """
        returns the parameter schemas by name
        """

    def __repr__(self):
        """
        :rtype: str
        """
        return "<{classname}>".format(
            classname=self.__class__.__name__
        )  # pragma: no cover

    @abc.abstractmethod
    def resolve_form(self, form_data):
        """Resolve cases where form parameters are provided multiple times."""

    @abc.abstractmethod
    def resolve_query(self, query_data):
        """Resolve cases where query parameters are provided multiple times."""

    @abc.abstractmethod
    def resolve_path(self, path):
        """Resolve cases where path parameters include lists"""

    @abc.abstractmethod
    def _resolve_param_duplicates(self, values, param_defn, _in):
        """Resolve cases where query parameters are provided multiple times.
        For example, if the query string is '?a=1,2,3&a=4,5,6' the value of
        `a` could be "4,5,6", or "1,2,3" or "1,2,3,4,5,6" depending on the
        implementation.
        """

    @abc.abstractmethod
    def _split(self, value, param_defn, _in):
        """
        takes a string, a parameter definition, and a parameter type
        and returns an array that has been constructed according to
        the parameter definition.
        """

    def resolve_params(self, params, _in):
        """
        takes a dict of parameters, and resolves the values into
        the correct array type handling duplicate values, and splitting
        based on the collectionFormat defined in the spec.
        """
        resolved_param = {}
        for k, values in params.items():
            param_defn = self.param_defns.get(k)
            param_schema = self.param_schemas.get(k)

            if not (param_defn or param_schema):
                # rely on validation
                resolved_param[k] = values
                continue

            if _in == "path":
                # multiple values in a path is impossible
                values = [values]

            if param_schema and param_schema["type"] == "array":
                # resolve variable re-assignment, handle explode
                values = self._resolve_param_duplicates(values, param_defn, _in)
                # handle array styles
                resolved_param[k] = self._split(values, param_defn, _in)
            else:
                resolved_param[k] = values[-1]

            try:
                resolved_param[k] = coerce_type(
                    param_defn, resolved_param[k], "parameter", k
                )
            except TypeValidationError:
                pass

        return resolved_param


class OpenAPIURIParser(AbstractURIParser):
    style_defaults = {
        "path": "simple",
        "header": "simple",
        "query": "form",
        "cookie": "form",
        "form": "form",
    }

    @property
    def param_defns(self):
        return self._param_defns

    @property
    def form_defns(self):
        return {k: v for k, v in self._body_schema.get("properties", {}).items()}

    @property
    def param_schemas(self):
        return {k: v.get("schema", {}) for k, v in self.param_defns.items()}

    def resolve_form(self, form_data):
        if self._body_schema is None or self._body_schema.get("type") != "object":
            return form_data
        for k in form_data:
            encoding = self._body_encoding.get(k, {"style": "form"})
            defn = self.form_defns.get(k, {})
            # TODO support more form encoding styles
            form_data[k] = self._resolve_param_duplicates(
                form_data[k], encoding, "form"
            )
            if "contentType" in encoding and all_json([encoding.get("contentType")]):
                form_data[k] = json.loads(form_data[k])
            elif defn and defn["type"] == "array":
                form_data[k] = self._split(form_data[k], encoding, "form")
            form_data[k] = coerce_type(defn, form_data[k], "requestBody", k)
        return form_data

    def _make_deep_object(self, k, v):
        """consumes keys, value pairs like (a[foo][bar], "baz")
        returns (a, {"foo": {"bar": "baz"}}}, is_deep_object)
        """
        root_key = None
        if k in self.param_schemas.keys():
            return k, v, False
        else:
            for key in self.param_schemas.keys():
                if k.startswith(key) and "[" in k:
                    root_key = key.replace(k, "")

        if not root_key:
            root_key = k.split("[", 1)[0]
            if k == root_key:
                return k, v, False

        if not self._is_deep_object_style_param(root_key):
            return k, v, False

        key_path = re.findall(r"\[([^\[\]]*)\]", k)
        root = prev = node = {}
        for k in key_path:
            node[k] = {}
            prev = node
            node = node[k]
        prev[k] = v[0]
        return root_key, [root], True

    def _is_deep_object_style_param(self, param_name):
        default_style = self.style_defaults["query"]
        style = self.param_defns.get(param_name, {}).get("style", default_style)
        return style == "deepObject"

    def _preprocess_deep_objects(self, query_data):
        """deep objects provide a way of rendering nested objects using query
        parameters.
        """
        deep = [self._make_deep_object(k, v) for k, v in query_data.items()]
        root_keys = [k for k, v, is_deep_object in deep]
        ret = dict.fromkeys(root_keys, [{}])
        for k, v, is_deep_object in deep:
            if is_deep_object:
                ret[k] = [deep_merge(v[0], ret[k][0])]
            else:
                ret[k] = v
        return ret

    def resolve_query(self, query_data):
        query_data = self._preprocess_deep_objects(query_data)
        return self.resolve_params(query_data, "query")

    def resolve_path(self, path_data):
        return self.resolve_params(path_data, "path")

    @staticmethod
    def _resolve_param_duplicates(values, param_defn, _in):
        """Resolve cases where query parameters are provided multiple times.
        The default behavior is to use the first-defined value.
        For example, if the query string is '?a=1,2,3&a=4,5,6' the value of
        `a` would be "4,5,6".
        However, if 'explode' is 'True' then the duplicate values
        are concatenated together and `a` would be "1,2,3,4,5,6".
        """
        default_style = OpenAPIURIParser.style_defaults[_in]
        style = param_defn.get("style", default_style)
        delimiter = QUERY_STRING_DELIMITERS.get(style, ",")
        is_form = style == "form"
        explode = param_defn.get("explode", is_form)
        if explode:
            return delimiter.join(values)

        # default to last defined value
        return values[-1]

    @staticmethod
    def _split(value, param_defn, _in):
        default_style = OpenAPIURIParser.style_defaults[_in]
        style = param_defn.get("style", default_style)
        delimiter = QUERY_STRING_DELIMITERS.get(style, ",")
        return value.split(delimiter)


class Swagger2URIParser(AbstractURIParser):
    """
    Adheres to the Swagger2 spec,
    Assumes that the last defined query parameter should be used.
    """

    parsable_parameters = ["query", "path", "formData"]

    @property
    def param_defns(self):
        return self._param_defns

    @property
    def param_schemas(self):
        return self._param_defns  # swagger2 conflates defn and schema

    def resolve_form(self, form_data):
        return self.resolve_params(form_data, "form")

    def resolve_query(self, query_data):
        return self.resolve_params(query_data, "query")

    def resolve_path(self, path_data):
        return self.resolve_params(path_data, "path")

    @staticmethod
    def _resolve_param_duplicates(values, param_defn, _in):
        """Resolve cases where query parameters are provided multiple times.
        The default behavior is to use the first-defined value.
        For example, if the query string is '?a=1,2,3&a=4,5,6' the value of
        `a` would be "4,5,6".
        However, if 'collectionFormat' is 'multi' then the duplicate values
        are concatenated together and `a` would be "1,2,3,4,5,6".
        """
        if param_defn.get("collectionFormat") == "multi":
            return ",".join(values)
        # default to last defined value
        return values[-1]

    @staticmethod
    def _split(value, param_defn, _in):
        if param_defn.get("collectionFormat") == "pipes":
            return value.split("|")
        return value.split(",")


class FirstValueURIParser(Swagger2URIParser):
    """
    Adheres to the Swagger2 spec
    Assumes that the first defined query parameter should be used
    """

    @staticmethod
    def _resolve_param_duplicates(values, param_defn, _in):
        """Resolve cases where query parameters are provided multiple times.
        The default behavior is to use the first-defined value.
        For example, if the query string is '?a=1,2,3&a=4,5,6' the value of
        `a` would be "1,2,3".
        However, if 'collectionFormat' is 'multi' then the duplicate values
        are concatenated together and `a` would be "1,2,3,4,5,6".
        """
        if param_defn.get("collectionFormat") == "multi":
            return ",".join(values)
        # default to first defined value
        return values[0]


class AlwaysMultiURIParser(Swagger2URIParser):
    """
    Does not adhere to the Swagger2 spec, but is backwards compatible with
    connexion behavior in version 1.4.2
    """

    @staticmethod
    def _resolve_param_duplicates(values, param_defn, _in):
        """Resolve cases where query parameters are provided multiple times.
        The default behavior is to join all provided parameters together.
        For example, if the query string is '?a=1,2,3&a=4,5,6' the value of
        `a` would be "1,2,3,4,5,6".
        """
        if param_defn.get("collectionFormat") == "pipes":
            return "|".join(values)
        return ",".join(values)



import functools
import re
import typing as t

import starlette.convertors
from starlette.responses import JSONResponse as StarletteJSONResponse
from starlette.responses import Response as StarletteResponse
from starlette.types import Receive, Scope

from connexion.frameworks.abstract import Framework
from connexion.lifecycle import ConnexionRequest
from connexion.uri_parsing import AbstractURIParser


class Starlette(Framework):
    @staticmethod
    def is_framework_response(response: t.Any) -> bool:
        return isinstance(response, StarletteResponse)

    @classmethod
    def connexion_to_framework_response(cls, response):
        return cls.build_response(
            status_code=response.status_code,
            content_type=response.content_type,
            headers=response.headers,
            data=response.body,
        )

    @classmethod
    def build_response(
        cls,
        data: t.Any,
        *,
        content_type: t.Optional[str] = None,
        headers: t.Optional[dict] = None,
        status_code: int = None,
    ):
        if isinstance(data, dict) or isinstance(data, list):
            response_cls = StarletteJSONResponse
        else:
            response_cls = StarletteResponse

        return response_cls(
            content=data,
            status_code=status_code,
            media_type=content_type,
            headers=headers,
        )

    @staticmethod
    def get_request(*, scope: Scope, receive: Receive, uri_parser: AbstractURIParser, **kwargs) -> ConnexionRequest:  # type: ignore
        return ConnexionRequest(scope, receive, uri_parser=uri_parser)


PATH_PARAMETER = re.compile(r"\{([^}]*)\}")
PATH_PARAMETER_CONVERTERS = {"integer": "int", "number": "float", "path": "path"}


def convert_path_parameter(match, types):
    name = match.group(1)
    swagger_type = types.get(name)
    converter = PATH_PARAMETER_CONVERTERS.get(swagger_type)
    return f'{{{name.replace("-", "_")}{":" if converter else ""}{converter or ""}}}'


def starlettify_path(swagger_path, types=None):
    """
    Convert swagger path templates to flask path templates

    :type swagger_path: str
    :type types: dict
    :rtype: str

    >>> starlettify_path('/foo-bar/{my-param}')
    '/foo-bar/{my_param}'

    >>> starlettify_path('/foo/{someint}', {'someint': 'int'})
    '/foo/{someint:int}'
    """
    if types is None:
        types = {}
    convert_match = functools.partial(convert_path_parameter, types=types)
    return PATH_PARAMETER.sub(convert_match, swagger_path)


class FloatConverter(starlette.convertors.FloatConvertor):
    """Starlette converter for OpenAPI number type"""

    regex = r"[+-]?[0-9]*(\.[0-9]*)?"


class IntegerConverter(starlette.convertors.IntegerConvertor):
    """Starlette converter for OpenAPI integer type"""

    regex = r"[+-]?[0-9]+"



"""
This module defines a FlaskApp, a Connexion application to wrap a Flask application.
"""
import functools
import pathlib
import typing as t

import flask
import starlette.exceptions
import werkzeug.exceptions
from a2wsgi import WSGIMiddleware
from flask import Response as FlaskResponse
from starlette.types import Receive, Scope, Send

from connexion.apps.abstract import AbstractApp
from connexion.decorators import FlaskDecorator
from connexion.exceptions import ResolverError
from connexion.frameworks import flask as flask_utils
from connexion.jsonifier import Jsonifier
from connexion.lifecycle import ConnexionRequest, ConnexionResponse
from connexion.middleware.abstract import AbstractRoutingAPI, SpecMiddleware
from connexion.middleware.lifespan import Lifespan
from connexion.operations import AbstractOperation
from connexion.options import SwaggerUIOptions
from connexion.resolver import Resolver
from connexion.types import MaybeAwaitable, WSGIApp
from connexion.uri_parsing import AbstractURIParser


class FlaskOperation:
    def __init__(
        self,
        fn: t.Callable,
        jsonifier: Jsonifier,
        operation_id: str,
        pythonic_params: bool,
    ) -> None:
        self._fn = fn
        self.jsonifier = jsonifier
        self.operation_id = operation_id
        self.pythonic_params = pythonic_params
        functools.update_wrapper(self, fn)

    @classmethod
    def from_operation(
        cls,
        operation: AbstractOperation,
        *,
        pythonic_params: bool,
        jsonifier: Jsonifier,
    ) -> "FlaskOperation":
        return cls(
            fn=operation.function,
            jsonifier=jsonifier,
            operation_id=operation.operation_id,
            pythonic_params=pythonic_params,
        )

    @property
    def fn(self) -> t.Callable:
        decorator = FlaskDecorator(
            pythonic_params=self.pythonic_params,
            jsonifier=self.jsonifier,
        )
        return decorator(self._fn)

    def __call__(self, *args, **kwargs) -> FlaskResponse:
        return self.fn(*args, **kwargs)


class FlaskApi(AbstractRoutingAPI):
    def __init__(
        self, *args, jsonifier: t.Optional[Jsonifier] = None, **kwargs
    ) -> None:
        self.jsonifier = jsonifier or Jsonifier(flask.json, indent=2)
        super().__init__(*args, **kwargs)

    def _set_base_path(self, base_path: t.Optional[str] = None) -> None:
        super()._set_base_path(base_path)
        self._set_blueprint()

    def _set_blueprint(self):
        endpoint = flask_utils.flaskify_endpoint(self.base_path) or "/"
        self.blueprint = flask.Blueprint(
            endpoint,
            __name__,
            url_prefix=self.base_path,
        )

    def _add_resolver_error_handler(self, method: str, path: str, err: ResolverError):
        pass

    def make_operation(self, operation):
        return FlaskOperation.from_operation(
            operation, pythonic_params=self.pythonic_params, jsonifier=self.jsonifier
        )

    @staticmethod
    def _framework_path_and_name(
        operation: AbstractOperation, path: str
    ) -> t.Tuple[str, str]:
        flask_path = flask_utils.flaskify_path(
            path, operation.get_path_parameter_types()
        )
        endpoint_name = flask_utils.flaskify_endpoint(
            operation.operation_id, operation.randomize_endpoint
        )
        return flask_path, endpoint_name

    def _add_operation_internal(
        self,
        method: str,
        path: str,
        operation: t.Callable,
        name: t.Optional[str] = None,
    ) -> None:
        self.blueprint.add_url_rule(path, name, operation, methods=[method])

    def add_url_rule(
        self,
        rule,
        endpoint: t.Optional[str] = None,
        view_func: t.Optional[t.Callable] = None,
        **options,
    ):
        return self.blueprint.add_url_rule(rule, endpoint, view_func, **options)


class FlaskASGIApp(SpecMiddleware):
    def __init__(self, import_name, server_args: dict, **kwargs):
        self.app = flask.Flask(import_name, **server_args)
        self.app.json = flask_utils.FlaskJSONProvider(self.app)
        self.app.url_map.converters["float"] = flask_utils.NumberConverter
        self.app.url_map.converters["int"] = flask_utils.IntegerConverter

        # Propagate Errors so we can handle them in the middleware
        self.app.config["PROPAGATE_EXCEPTIONS"] = True
        self.app.config["TRAP_BAD_REQUEST_ERRORS"] = True
        self.app.config["TRAP_HTTP_EXCEPTIONS"] = True

        self.asgi_app = WSGIMiddleware(self.app.wsgi_app)

    def add_api(self, specification, *, name: t.Optional[str] = None, **kwargs):
        api = FlaskApi(specification, **kwargs)

        if name is not None:
            self.app.register_blueprint(api.blueprint, name=name)
        else:
            self.app.register_blueprint(api.blueprint)

        return api

    def add_url_rule(
        self,
        rule,
        endpoint: t.Optional[str] = None,
        view_func: t.Optional[t.Callable] = None,
        **options,
    ):
        return self.app.add_url_rule(rule, endpoint, view_func, **options)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        return await self.asgi_app(scope, receive, send)


class FlaskApp(AbstractApp):
    """Connexion Application based on ConnexionMiddleware wrapping a Flask application."""

    _middleware_app: FlaskASGIApp

    def __init__(
        self,
        import_name: str,
        *,
        lifespan: t.Optional[Lifespan] = None,
        middlewares: t.Optional[list] = None,
        server_args: t.Optional[dict] = None,
        specification_dir: t.Union[pathlib.Path, str] = "",
        arguments: t.Optional[dict] = None,
        auth_all_paths: t.Optional[bool] = None,
        jsonifier: t.Optional[Jsonifier] = None,
        pythonic_params: t.Optional[bool] = None,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = None,
        resolver_error: t.Optional[int] = None,
        strict_validation: t.Optional[bool] = None,
        swagger_ui_options: t.Optional[SwaggerUIOptions] = None,
        uri_parser_class: t.Optional[AbstractURIParser] = None,
        validate_responses: t.Optional[bool] = None,
        validator_map: t.Optional[dict] = None,
        security_map: t.Optional[dict] = None,
    ):
        """
        :param import_name: The name of the package or module that this object belongs to. If you
            are using a single module, __name__ is always the correct value. If you however are
            using a package, it’s usually recommended to hardcode the name of your package there.
        :param lifespan: A lifespan context function, which can be used to perform startup and
            shutdown tasks.
        :param middlewares: The list of middlewares to wrap around the application. Defaults to
            :obj:`middleware.main.ConnexionMiddleware.default_middlewares`
        :param server_args: Arguments to pass to the Flask application.
        :param specification_dir: The directory holding the specification(s). The provided path
            should either be absolute or relative to the root path of the application. Defaults to
            the root path.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate all paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: Instance of :class:`options.SwaggerUIOptions` with
            configuration options for the swagger ui.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`.
        :param security_map: A dictionary of security handlers to use. Defaults to
            :obj:`security.SECURITY_HANDLERS`
        """
        self._middleware_app = FlaskASGIApp(import_name, server_args or {})

        super().__init__(
            import_name,
            lifespan=lifespan,
            middlewares=middlewares,
            specification_dir=specification_dir,
            arguments=arguments,
            auth_all_paths=auth_all_paths,
            jsonifier=jsonifier,
            pythonic_params=pythonic_params,
            resolver=resolver,
            resolver_error=resolver_error,
            strict_validation=strict_validation,
            swagger_ui_options=swagger_ui_options,
            uri_parser_class=uri_parser_class,
            validate_responses=validate_responses,
            validator_map=validator_map,
            security_map=security_map,
        )

        self.app = self._middleware_app.app
        self.app.register_error_handler(
            werkzeug.exceptions.HTTPException, self._http_exception
        )

    def _http_exception(self, exc: werkzeug.exceptions.HTTPException):
        """Reraise werkzeug HTTPExceptions as starlette HTTPExceptions"""
        raise starlette.exceptions.HTTPException(exc.code, detail=exc.description)

    def add_url_rule(
        self,
        rule,
        endpoint: t.Optional[str] = None,
        view_func: t.Optional[t.Callable] = None,
        **options,
    ):
        self._middleware_app.add_url_rule(
            rule, endpoint=endpoint, view_func=view_func, **options
        )

    def add_error_handler(
        self,
        code_or_exception: t.Union[int, t.Type[Exception]],
        function: t.Callable[
            [ConnexionRequest, Exception], MaybeAwaitable[ConnexionResponse]
        ],
    ) -> None:
        self.middleware.add_error_handler(code_or_exception, function)

    def add_wsgi_middleware(
        self, middleware: t.Type[WSGIApp], **options: t.Any
    ) -> None:
        """Wrap the underlying Flask application with a WSGI middleware. Note that it will only be
        called at the end of the middleware stack. Middleware that needs to act sooner, needs to
        be added as ASGI middleware instead.

        Adding multiple middleware using this method wraps each middleware around the previous one.

        :param middleware: Middleware class to add
        :param options: Options to pass to the middleware_class on initialization
        """
        self._middleware_app.asgi_app.app = middleware(
            self._middleware_app.asgi_app.app, **options  # type: ignore
        )



"""
This module defines Python interfaces for OpenAPI specifications.
"""

import abc
import copy
import json
import os
import pathlib
import pkgutil
import typing as t
from collections.abc import Mapping
from urllib.parse import urlsplit

import jinja2
import jsonschema
import yaml
from jsonschema import Draft4Validator
from jsonschema.validators import extend as extend_validator

from .exceptions import InvalidSpecification
from .json_schema import NullableTypeValidator, URLHandler, resolve_refs
from .operations import AbstractOperation, OpenAPIOperation, Swagger2Operation
from .utils import deep_get

validate_properties = Draft4Validator.VALIDATORS["properties"]


def create_spec_validator(spec: dict) -> Draft4Validator:
    """Create a Validator to validate an OpenAPI spec against the OpenAPI schema.

    :param spec: specification to validate
    """
    # Create an instance validator, which validates defaults against the spec itself instead of
    # against the OpenAPI schema.
    InstanceValidator = extend_validator(
        Draft4Validator, {"type": NullableTypeValidator}
    )
    instance_validator = InstanceValidator(spec)

    def validate_defaults(validator, properties, instance, schema):
        """Validation function to validate the `properties` subschema, enforcing each default
        value validates against the schema in which it resides.
        """
        valid = True
        for error in validate_properties(validator, properties, instance, schema):
            valid = False
            yield error

        # Validate default only when the subschema has validated successfully
        if not valid:
            return
        if isinstance(instance, dict) and "default" in instance:
            for error in instance_validator.evolve(schema=instance).iter_errors(
                instance["default"]
            ):
                yield error

    SpecValidator = extend_validator(Draft4Validator, {"properties": validate_defaults})
    return SpecValidator


NO_SPEC_VERSION_ERR_MSG = """Unable to get the spec version.
You are missing either '"swagger": "2.0"' or '"openapi": "3.0.0"'
from the top level of your spec."""


def canonical_base_path(base_path):
    """
    Make given "basePath" a canonical base URL which can be prepended to paths starting with "/".
    """
    return base_path.rstrip("/")


class Specification(Mapping):

    operation_cls: t.Type[AbstractOperation]

    def __init__(self, raw_spec, *, base_uri=""):
        self._raw_spec = copy.deepcopy(raw_spec)
        self._set_defaults(raw_spec)
        self._validate_spec(raw_spec)
        self._spec = resolve_refs(raw_spec, base_uri=base_uri)
        self._base_uri = base_uri

    @classmethod
    @abc.abstractmethod
    def _set_defaults(cls, spec):
        """set some default values in the spec"""

    @classmethod
    def _validate_spec(cls, spec):
        """validate spec against schema"""
        try:
            OpenApiValidator = create_spec_validator(spec)
            validator = OpenApiValidator(cls.openapi_schema)
            validator.validate(spec)
        except jsonschema.exceptions.ValidationError as e:
            raise InvalidSpecification.create_from(e)

    def get_path_params(self, path):
        return deep_get(self._spec, ["paths", path]).get("parameters", [])

    def get_operation(self, path, method):
        return deep_get(self._spec, ["paths", path, method])

    @property
    def raw(self):
        return self._raw_spec

    @property
    def spec(self):
        return self._spec

    @property
    def version(self):
        return self._get_spec_version(self._spec)

    @property
    def security(self):
        return self._spec.get("security")

    @property
    @abc.abstractmethod
    def security_schemes(self):
        raise NotImplementedError

    def __getitem__(self, k):
        return self._spec[k]

    def __iter__(self):
        return self._spec.__iter__()

    def __len__(self):
        return self._spec.__len__()

    @staticmethod
    def _load_spec_from_file(arguments, specification):
        """
        Loads a YAML specification file, optionally rendering it with Jinja2.

        :param arguments: passed to Jinja2 renderer
        :param specification: path to specification
        """
        arguments = arguments or {}

        with specification.open(mode="rb") as openapi_yaml:
            contents = openapi_yaml.read()
            try:
                openapi_template = contents.decode()
            except UnicodeDecodeError:
                openapi_template = contents.decode("utf-8", "replace")

            openapi_string = jinja2.Template(openapi_template).render(**arguments)
            return yaml.safe_load(openapi_string)

    @classmethod
    def from_file(cls, spec, *, arguments=None, base_uri=""):
        """
        Takes in a path to a YAML file, and returns a Specification
        """
        specification_path = pathlib.Path(spec)
        spec = cls._load_spec_from_file(arguments, specification_path)
        return cls.from_dict(spec, base_uri=base_uri)

    @classmethod
    def from_url(cls, spec, *, base_uri=""):
        """
        Takes in a path to a YAML file, and returns a Specification
        """
        spec = URLHandler()(spec)
        return cls.from_dict(spec, base_uri=base_uri)

    @staticmethod
    def _get_spec_version(spec):
        try:
            version_string = spec.get("openapi") or spec.get("swagger")
        except AttributeError:
            raise InvalidSpecification(NO_SPEC_VERSION_ERR_MSG)
        if version_string is None:
            raise InvalidSpecification(NO_SPEC_VERSION_ERR_MSG)
        try:
            version_tuple = tuple(map(int, version_string.split(".")))
        except TypeError:
            err = (
                "Unable to convert version string to semantic version tuple: "
                "{version_string}."
            )
            err = err.format(version_string=version_string)
            raise InvalidSpecification(err)
        return version_tuple

    @classmethod
    def from_dict(cls, spec, *, base_uri=""):
        """
        Takes in a dictionary, and returns a Specification
        """

        def enforce_string_keys(obj):
            # YAML supports integer keys, but JSON does not
            if isinstance(obj, dict):
                return {str(k): enforce_string_keys(v) for k, v in obj.items()}
            return obj

        spec = enforce_string_keys(spec)
        version = cls._get_spec_version(spec)
        if version < (3, 0, 0):
            return Swagger2Specification(spec, base_uri=base_uri)
        return OpenAPISpecification(spec, base_uri=base_uri)

    def clone(self):
        return type(self)(copy.deepcopy(self._raw_spec), base_uri=self._base_uri)

    @classmethod
    def load(cls, spec, *, arguments=None):
        if isinstance(spec, str) and (
            spec.startswith("http://") or spec.startswith("https://")
        ):
            return cls.from_url(spec)
        if not isinstance(spec, dict):
            base_uri = f"{pathlib.Path(spec).parent}{os.sep}"
            return cls.from_file(spec, arguments=arguments, base_uri=base_uri)
        return cls.from_dict(spec)

    def with_base_path(self, base_path):
        new_spec = self.clone()
        new_spec.base_path = base_path
        return new_spec

    @property
    @abc.abstractmethod
    def base_path(self):
        pass

    @base_path.setter
    @abc.abstractmethod
    def base_path(self, base_path):
        pass


class Swagger2Specification(Specification):
    """Python interface for a Swagger 2 specification."""

    yaml_name = "swagger.yaml"
    operation_cls = Swagger2Operation

    openapi_schema = json.loads(
        pkgutil.get_data("connexion", "resources/schemas/v2.0/schema.json")  # type: ignore
    )

    @classmethod
    def _set_defaults(cls, spec):
        spec.setdefault("produces", [])
        spec.setdefault("consumes", ["application/json"])
        spec.setdefault("definitions", {})
        spec.setdefault("parameters", {})
        spec.setdefault("responses", {})

    @property
    def produces(self):
        return self._spec["produces"]

    @property
    def consumes(self):
        return self._spec["consumes"]

    @property
    def definitions(self):
        return self._spec["definitions"]

    @property
    def parameter_definitions(self):
        return self._spec["parameters"]

    @property
    def response_definitions(self):
        return self._spec["responses"]

    @property
    def security_schemes(self):
        return self._spec.get("securityDefinitions", {})

    @property
    def base_path(self):
        return canonical_base_path(self._spec.get("basePath", ""))

    @base_path.setter
    def base_path(self, base_path):
        base_path = canonical_base_path(base_path)
        self._raw_spec["basePath"] = base_path
        self._spec["basePath"] = base_path


class OpenAPISpecification(Specification):
    """Python interface for an OpenAPI 3 specification."""

    yaml_name = "openapi.yaml"
    operation_cls = OpenAPIOperation

    openapi_schema = json.loads(
        pkgutil.get_data("connexion", "resources/schemas/v3.0/schema.json")  # type: ignore
    )

    @classmethod
    def _set_defaults(cls, spec):
        spec.setdefault("components", {})

    @property
    def security_schemes(self):
        return self._spec["components"].get("securitySchemes", {})

    @property
    def components(self):
        return self._spec["components"]

    @property
    def base_path(self):
        servers = self._spec.get("servers", [])
        try:
            # assume we're the first server in list
            server = copy.deepcopy(servers[0])
            server_vars = server.pop("variables", {})
            server["url"] = server["url"].format(
                **{k: v["default"] for k, v in server_vars.items()}
            )
            base_path = urlsplit(server["url"]).path
        except IndexError:
            base_path = ""
        return canonical_base_path(base_path)

    @base_path.setter
    def base_path(self, base_path):
        base_path = canonical_base_path(base_path)
        user_servers = [{"url": base_path}]
        self._raw_spec["servers"] = user_servers
        self._spec["servers"] = user_servers



import logging
import typing as t

from starlette.middleware.errors import (
    ServerErrorMiddleware as StarletteServerErrorMiddleware,
)
from starlette.types import ASGIApp

from connexion.exceptions import InternalServerError
from connexion.lifecycle import ConnexionRequest, ConnexionResponse
from connexion.middleware.exceptions import connexion_wrapper
from connexion.types import MaybeAwaitable

logger = logging.getLogger(__name__)


class ServerErrorMiddleware(StarletteServerErrorMiddleware):
    """Subclass of starlette ServerErrorMiddleware to change handling of Unhandled Server
    exceptions to existing connexion behavior."""

    def __init__(
        self,
        next_app: ASGIApp,
        handler: t.Optional[
            t.Callable[[ConnexionRequest, Exception], MaybeAwaitable[ConnexionResponse]]
        ] = None,
    ):
        handler = connexion_wrapper(handler) if handler else None
        super().__init__(next_app, handler=handler)

    @staticmethod
    @connexion_wrapper
    def error_response(_request: ConnexionRequest, exc: Exception) -> ConnexionResponse:
        """Default handler for any unhandled Exception"""
        logger.error("%r", exc, exc_info=exc)
        return InternalServerError().to_problem()



import abc
import collections.abc
import functools
import logging
import types
import typing as t
from enum import Enum

from connexion import utils
from connexion.context import operation
from connexion.datastructures import NoContent
from connexion.exceptions import NonConformingResponseHeaders
from connexion.frameworks.abstract import Framework
from connexion.lifecycle import ConnexionResponse

logger = logging.getLogger(__name__)


class BaseResponseDecorator:
    def __init__(self, *, framework: t.Type[Framework], jsonifier):
        self.framework = framework
        self.jsonifier = jsonifier

    @abc.abstractmethod
    def __call__(self, function: t.Callable) -> t.Callable:
        raise NotImplementedError

    def build_framework_response(self, handler_response):
        data, status_code, headers = self._unpack_handler_response(handler_response)
        content_type = self._infer_content_type(data, headers)
        if not self.framework.is_framework_response(data):
            data = self._serialize_data(data, content_type=content_type)
            status_code = status_code or self._infer_status_code(data)
            headers = self._update_headers(headers, content_type=content_type)
        return self.framework.build_response(
            data, content_type=content_type, status_code=status_code, headers=headers
        )

    @staticmethod
    def _infer_content_type(data: t.Any, headers: dict) -> t.Optional[str]:
        """Infer the response content type from the returned data, headers and operation spec.

        :param data: Response data
        :param headers: Headers returned by the handler.

        :return: Inferred content type

        :raises: NonConformingResponseHeaders if content type cannot be deducted.
        """
        content_type = utils.extract_content_type(headers)

        # TODO: don't default
        produces = list(set(operation.produces))
        if data is not None and not produces:
            produces = ["application/json"]

        if content_type:
            if content_type not in produces:
                raise NonConformingResponseHeaders(
                    f"Returned content type ({content_type}) is not defined in operation spec "
                    f"({operation.produces})."
                )
        else:
            if not produces:
                # Produces can be empty/ for empty responses
                pass
            elif len(produces) == 1:
                content_type = produces[0]
            else:
                if isinstance(data, str):
                    for produced_content_type in produces:
                        if "text/plain" in produced_content_type:
                            content_type = produced_content_type
                elif isinstance(data, bytes) or isinstance(
                    data, (types.GeneratorType, collections.abc.Iterator)
                ):
                    for produced_content_type in produces:
                        if "application/octet-stream" in produced_content_type:
                            content_type = produced_content_type

                if content_type is None:
                    raise NonConformingResponseHeaders(
                        "Multiple response content types are defined in the operation spec, but "
                        "the handler response did not specify which one to return."
                    )

        return content_type

    def _serialize_data(self, data: t.Any, *, content_type: str) -> t.Any:
        """Serialize the data based on the content type."""
        if data is None or data is NoContent:
            return None
        # TODO: encode responses
        mime_type, _ = utils.split_content_type(content_type)
        if utils.is_json_mimetype(mime_type):
            return self.jsonifier.dumps(data)
        return data

    @staticmethod
    def _infer_status_code(data: t.Any) -> int:
        """Infer the status code from the returned data."""
        if data is None:
            return 204
        return 200

    @staticmethod
    def _update_headers(
        headers: t.Dict[str, str], *, content_type: str
    ) -> t.Dict[str, str]:
        # Check if Content-Type is in headers, taking into account case-insensitivity
        for key, value in headers.items():
            if key.lower() == "content-type":
                return headers

        if content_type:
            headers["Content-Type"] = content_type
        return headers

    @staticmethod
    def _unpack_handler_response(
        handler_response: t.Union[str, bytes, dict, list, tuple]
    ) -> t.Tuple[t.Union[str, bytes, dict, list, None], t.Optional[int], dict]:
        """Unpack the handler response into data, status_code and headers.

        :param handler_response: The response returned from the handler function if it was not a
            response class.

        :return: A tuple of data, status_code and headers
        """
        data, status_code, headers = None, None, {}

        if not isinstance(handler_response, tuple):
            data = handler_response

        elif len(handler_response) == 1:
            (data,) = handler_response

        elif len(handler_response) == 2:
            data, status_code_or_headers = handler_response
            if isinstance(status_code_or_headers, int):
                # Extra int call because of int subclasses such as http.HTTPStatus (IntEnum)
                status_code = int(status_code_or_headers)
            elif isinstance(status_code_or_headers, Enum) and isinstance(
                status_code_or_headers.value, int
            ):
                status_code = status_code_or_headers.value
            else:
                headers = status_code_or_headers

        elif len(handler_response) == 3:
            data, status_code, headers = handler_response

        else:
            raise TypeError(
                "The view function did not return a valid response tuple."
                " The tuple must have the form (body), (body, status, headers),"
                " (body, status), or (body, headers)."
            )

        return data, status_code, headers


class SyncResponseDecorator(BaseResponseDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            """
            This method converts a handler response to a framework response.
            The handler response can be a ConnexionResponse, a framework response, a tuple or an
            object.
            """
            handler_response = function(*args, **kwargs)
            if self.framework.is_framework_response(handler_response):
                return handler_response
            elif isinstance(handler_response, ConnexionResponse):
                return self.framework.connexion_to_framework_response(handler_response)
            else:
                return self.build_framework_response(handler_response)

        return wrapper


class AsyncResponseDecorator(BaseResponseDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        @functools.wraps(function)
        async def wrapper(*args, **kwargs):
            """
            This method converts a handler response to a framework response.
            The handler response can be a ConnexionResponse, a framework response, a tuple or an
            object.
            """
            handler_response = await function(*args, **kwargs)
            if self.framework.is_framework_response(handler_response):
                return handler_response
            elif isinstance(handler_response, ConnexionResponse):
                return self.framework.connexion_to_framework_response(handler_response)
            else:
                return self.build_framework_response(handler_response)

        return wrapper


class NoResponseDecorator(BaseResponseDecorator):
    """Dummy decorator to skip response serialization."""

    def __call__(self, function: t.Callable) -> t.Callable:
        return lambda request: function(request)
